"""
Useful functions for spillover effects estimation
"""

__author__ = """ Pablo Estrada pabloestradace@gmail.com """

import numpy as np
import pandas as pd
from scipy import sparse as spr
from scipy.stats import hypergeom


def adjacency_matrix(edges, directed=True, nodes=None):
    """
    Adjacency matrix and nodelist from edge list dataframe

    Parameters
    ----------
    edges     : pandas.core.frame.DataFrame
                Dataframe with j+1 columns
                Column 1 is the source node
                Column 2,3,j+1 are the target nodes
    directed  : bool
                Whether the graph is directed or not
    nodes     : array
                n x 1 array of the nodes order
    """
    # Transform edges to two columns of source and target nodes
    data = edges.iloc[:, 0:2].dropna()
    data.columns = [0, 1]
    for j in range(2, edges.shape[1]):
        data_j = edges.iloc[:, [0, j]].dropna()
        data_j.columns = [0, 1]
        data = pd.concat([data, data_j], ignore_index=True)
    # Get unique nodes
    nodes = edges.stack().unique() if nodes is None else nodes
    n = len(nodes)
    # Create mapping of nodes to indices
    nodes_map = {nodes[i]: str(i) for i in range(n)}
    data = data.replace(nodes_map)
    rows = data[0].astype(int)
    cols = data[1].astype(int)
    ones = np.ones(len(rows), np.uint32)
    A = spr.coo_matrix((ones, (rows, cols)), shape=(n, n))
    if not directed:
        A = A + A.T
    return A, nodes


def spillover_treatment(treatment, A, interaction=False):
    """
    Treatment matrix for spillover effects estimation

    Parameters
    ----------
    treatment   : array
                  n x 1 array of treatment assignment
    A           : array
                  n x n adjacency matrix
    interaction : bool
                  Whether to include the interaction of direct and spillover treatments
    """
    if interaction:
        spillover = ((A @ treatment) > 0) * 1
        return np.vstack([(1-treatment) * (1-spillover), 
                          (1-treatment) * spillover, 
                          treatment     * (1-spillover), 
                          treatment     * spillover]).T
    else:
        spillover = ((A @ treatment) > 0) * 1
        return np.vstack([1-spillover, spillover]).T


def spillover_pscore(A, n_treated, blocks=None, matrix=False):
    """
    Compute the propensity score of having at least one friend treated

    Parameters
    ----------
    A         : array
                n x n adjacency matrix
    n_treated : int
                Number of treated individuals in the block
    blocks    : array
                n x 1 array of block assignments, k unique blocks
    matrix    : bool
                Whether to return the matrix of propensity scores
    """
    n = A.shape[0]
    if blocks is None:
        # Protocol: all students are in the same block
        degree = A @ np.ones(n)
        pscore_spillover = 1 - hypergeom(n, n_treated, degree).pmf(0)
        pscore_direct = n_treated / n
        # pscore0_spillover = binom(degree, pscore0_direct).pmf(0)
    else:
        # Protocol: propensity score by blocks, e.g., classrooms
        unique_blocks = blocks.unique()
        # Each row is a vector giving the number of friends of each student that are in classroom k
        degree_by_block = np.vstack([A @ (blocks==k).values for k in unique_blocks])
        # 174 blocks (classrooms) of x students, 4 treated
        blocks_size = blocks.value_counts().loc[unique_blocks].values
        k = len(unique_blocks)
        p0_block = np.zeros((k, n))
        # Probability of having zero treated friends out of the x students in the k classroom
        for i in range(k):
            p0_block[i, :] = hypergeom(blocks_size[i], n_treated, degree_by_block[i, :]).pmf(0)
        pscore_spillover = 1 - p0_block.prod(axis=0)
        pscore_direct = [n_treated / blocks.value_counts().loc[i] for i in blocks]
    if matrix:
        return np.vstack([(1-pscore_direct) * (1-pscore_spillover), 
                          (1-pscore_direct) * pscore_spillover, 
                          pscore_direct     * (1-pscore_spillover), 
                          pscore_direct     * pscore_spillover]).T
    else:
        return np.vstack([1-pscore_spillover, pscore_spillover]).T


def kernel(A, bw):
    """
    Kernel matrix for covariance estimation

    Parameters
    ----------
    A         : array
                n x n adjacency matrix
    bw        : int
                Bandwidth to calculate kernel matrix
    """
    if spr.issparse(A):
        A = A.toarray()
    # Calculate shortest path distance matrix
    dist_matrix = spr.csgraph.dijkstra(csgraph=A, directed=False, unweighted=True)
    # Calculate kernel matrix
    weights = (dist_matrix <= bw) * 1
    # Check for negative eigenvalues
    eigenvalues, eigenvectors = np.linalg.eigh(weights)
    if eigenvalues[0] < 0:
        weights = eigenvectors @ np.diag(np.maximum(eigenvalues, 0)) @ eigenvectors.T
    return weights


"""
WLS Estimation of Spillover Effects
"""

__author__ = """ Pablo Estrada pabloestradace@gmail.com """

import numpy as np
import pandas as pd
from scipy import sparse as spr
from scipy.stats import norm


class WLS():
    
        """
        WLS estimation of spillover effects under a pre-specified exposure mapping.
    
        Parameters
        ----------
        y             : array
                        n x 1 array of dependent variable
        Z             : array
                        n x t array of exposure mapping
        pscore        : array
                        n x 1 array of propensity scores
        A             : array
                        n x n adjacency matrix
        X             : array
                        n x k array of covariates
        interaction   : bool
                        Whether to include the interaction of Z and X
        bw            : int
                        Bandwidth for kernel matrix
    
        Attributes
        ----------
        params        : array
                        (t+k) x 1 array of WLS coefficients
        vcov          : array
                        Variance covariance matrix
        """
    
        def __init__(self,
                    name_y,
                    name_z,
                    name_pscore,
                    data,
                    kernel_weights=None,
                    name_x=None,
                    interaction=True,
                    subsample=None,
                    contrast='spillover'):

            # Outcome and treatment exposure
            y = data[name_y].values
            Z = data[name_z].values
            pscore = data[name_pscore].values
            # Standardize or create matrix X
            t = Z.shape[1]
            name_x = [name_x] if isinstance(name_x, str) else name_x
            X = data[name_x].values if name_x is not None else None
            if X is not None:
                X = (X - X.mean(axis=0)) # / X.std(axis=0)
                if interaction:
                    ZX = np.hstack([Z[:, i:i+1] * X for i in range(t)])
                    X = np.hstack((Z, ZX))
                else:
                    X = np.hstack((Z, X))
            else:
                X = Z.copy() if X is None else X
            # Kernel matrix
            n = Z.shape[0]
            weights = np.identity(n) if kernel_weights is None else kernel_weights
            # Filter by subsample of interest
            if subsample is not None:
                y = y[subsample]
                Z = Z[subsample]
                pscore = pscore[subsample]
                selection = selection[subsample]
                weights = weights[subsample,:][:,subsample]
                X = X[subsample] if X is not None else None
            # Check for propensity score outside (0.01, 0.99)
            valid = (np.sum(Z*pscore, axis=1) > 0.01) & (np.sum(Z*pscore, axis=1) < 0.99)
            drop_obs = np.sum(~valid)
            if drop_obs > 0:
                print('Warning: {} observations have propensity scores outside (0.01, 0.99)'.format(drop_obs))
                y = y[valid]
                Z = Z[valid]
                pscore = pscore[valid]
                weights = weights[valid,:][:,valid]
                X = X[valid] if X is not None else None
            # Weight with propensity score
            W = np.diag(1 / np.sum(Z*pscore, axis=1))
            # Fit WLS
            XWXi = sinv(X.T @ W @ X)
            beta = XWXi @ X.T @ W @ y
            # Variance
            e = np.diag(y - X @ beta)
            V = XWXi @ X.T @ W @ e @ weights @ e @ W @ X @ XWXi
            # Summary of results
            if t == 4:
                G = 1/2 * np.array([-1, -1, 1, 1]) if contrast is 'direct' else 1/2 * np.array([-1, 1, -1, 1])
            elif t == 2:
                G = np.array([-1, 1])
            else:
                raise ValueError('Contrast not available for T lenght = {}'.format(t))
            coef = np.insert(beta, 0, G @ beta[:t])
            se = np.insert(np.sqrt(V.diagonal()), 0, np.sqrt(G @ V[:t, :t] @ G.T))
            tval = coef / se
            pval = 2 * (1 - norm.cdf(np.abs(tval)))
            ci_low = coef - 1.96*se
            ci_up = coef + 1.96*se
            if name_x is None:
                name_vars = [contrast] + name_z
            else:
                if interaction:
                    name_vars = [contrast] + name_z + [zi + '*' + xi for zi in name_z for xi in name_x]
                else: 
                    name_vars = [contrast] + name_z + name_x
            df_results = pd.DataFrame({'coef': coef, 'se': se, 't-val': tval, 'p-val': pval,
                                       'ci-low': ci_low, 'ci-up': ci_up},
                                      index=name_vars)

            self.params = beta
            self.vcov = V
            self.summary = df_results



def sinv(A):
    """
    Find inverse of matrix A using numpy.linalg.solve
    Helpful for large matrices
    """
    if spr.issparse(A):
        n = A.shape[0]
        Ai = spr.linalg.spsolve(A.tocsc(), spr.identity(n, format='csc'))
    else:
        try:
            n = A.shape[0]
            Ai = np.linalg.solve(A, np.identity(n))
        except np.linalg.LinAlgError as err:
            if 'Singular matrix' in str(err):
                Ai = np.linalg.pinv(A)
            else:
                raise
    return Ai


"""
WLS Estimation of Spillover Bounds
"""

__author__ = """ Pablo Estrada pabloestradace@gmail.com """

import numpy as np
from scipy import sparse as spr
from .utils import kernel


class Bounds():
    
        """
        Bounds using WLS estimation of spillover effects under a pre-specified exposure mapping.
    
        Parameters
        ----------
        y             : array
                        ns x 1 array of dependent variable
        selection     : array
                        n x 1 array of selection variable
        Z             : array
                        n x t array of exposure mapping
        pscore        : array
                        n x t array of propensity scores
        group_exposed : array
                        t x 1 boolean array indicating the group exposed (1) or control (0)
        A             : array
                        n x n adjacency matrix
        bw            : int
                        Bandwidth for kernel matrix
    
        Attributes
        ----------
        params        : array
                        (t+k) x 1 array of WLS coefficients
        vcov          : array
                        Variance covariance matrix
        """
    
        def __init__(self,
                    y,
                    selection,
                    Z,
                    pscore,
                    group_exposed,
                    A=None,
                    subsample=None,
                    bw=2):
            
            # Kernel matrix
            n = Z.shape[0]
            weights = kernel(A, bw) if A is not None else np.identity(n)
            # Filter by subsample of interest
            if subsample is not None:
                y = y[subsample[selection]]
                Z = Z[subsample]
                pscore = pscore[subsample]
                selection = selection[subsample]
                weights = weights[subsample,:][:,subsample]
            # Check for propensity score outside (0.01, 0.99)
            valid = (np.sum(Z*pscore, axis=1) > 0.01) & (np.sum(Z*pscore, axis=1) < 0.99)
            drop_obs = np.sum(~valid)
            if drop_obs > 0:
                print('Warning: {} observations have propensity scores outside (0.01, 0.99)'.format(drop_obs))
                y = y[valid[selection]]
                Z = Z[valid]
                n = Z.shape[0]
                pscore = pscore[valid]
                weights = weights[valid,:][:,valid]
                selection = selection[valid]
            # Calculate trimming probability
            p_hat, group_trimmed, alpha_hat = trim_prob(selection, Z, pscore, group_exposed)
            n_list = ns_for_variance(selection, Z, pscore, group_exposed, group_trimmed)
            # Calculate trimming indicators
            Zs = Z[selection]
            pscore_s = pscore[selection]
            weights = weights[selection,:][:,selection]
            trim_bounds, quantile_bounds = trim_quantile(y, Zs, group_exposed, p_hat, group_trimmed)
            beta_bounds, V_bounds = [], []
            for i in range(2):
                trim = trim_bounds[i]
                quantile = quantile_bounds[i]
                y_trim = y[trim >= 0]
                Z_trim = Zs[trim >= 0]
                pscore_trim = pscore_s[trim >= 0]
                weights_trim = weights[trim >= 0, :][:, trim >= 0]
                # Weight with propensity score
                W = np.diag(1 / np.sum(Z_trim*pscore_trim, axis=1))
                # Fit WLS
                XWXi = sinv(Z_trim.T @ W @ Z_trim)
                beta = XWXi @ Z_trim.T @ W @ y_trim
                # Variance
                e = np.diag(y_trim - Z_trim @ beta)
                V = XWXi @ Z_trim.T @ W @ e @ weights_trim @ e @ W @ Z_trim @ XWXi
                v_y = (quantile - beta[group_exposed==group_trimmed])**2 * (1-p_hat)/p_hat / n_list[3-group_trimmed]
                v_q = ((quantile - beta[group_exposed==group_trimmed])**2) * ((p_hat-alpha_hat)/alpha_hat/n_list[0] + (1-alpha_hat)/alpha_hat/n_list[1])
                V_unadj = V[group_exposed==group_trimmed, group_exposed==group_trimmed]
                V[group_exposed==group_trimmed, group_exposed==group_trimmed] = V_unadj + v_y + v_q
                beta_bounds.append(beta)
                V_bounds.append(V)

            self.params_lb = beta_bounds[0]
            self.params_ub = beta_bounds[1]
            self.vcov_lb = V_bounds[0]
            self.vcov_ub = V_bounds[1]
            self.n = n
            self.p_hat = p_hat
            self.group_trimmed = group_trimmed
            self.alpha_hat = alpha_hat


def sinv(A):
    """
    Find inverse of matrix A using numpy.linalg.solve
    Helpful for large matrices
    """
    if spr.issparse(A):
        n = A.shape[0]
        Ai = spr.linalg.spsolve(A.tocsc(), spr.identity(n, format='csc'))
    else:
        try:
            n = A.shape[0]
            Ai = np.linalg.solve(A, np.identity(n))
        except np.linalg.LinAlgError as err:
            if 'Singular matrix' in str(err):
                Ai = np.linalg.pinv(A)
            else:
                raise
    return Ai


def trim_prob(S, Z, p, group):
    """
    Calculate the trimming probability q and the trimmed-group indicator g

    Parameters
    ----------
    S : array 
        n x 1 boolean array indicating the selected sample
    Z : array
        n x t exposure mapping matrix
    p : array
        n x t propensity score matrix
    group : array
        t x1 boolean array indicating the group exposed (1) or control (0)
    """
    Zs = Z[S]
    ps = p[S]
    W = 1 / np.sum(Z*p, axis=1)
    Ws = 1 / np.sum(Zs*ps, axis=1)
    qT = np.sum(Ws[Zs[:, group==1].sum(axis=1) == 1]) / np.sum(W[Z[:, group==1].sum(axis=1) == 1])
    qC = np.sum(Ws[Zs[:, group==0].sum(axis=1) == 1]) / np.sum(W[Z[:, group==0].sum(axis=1) == 1])
    # qT = np.sum(Zs[:, group==1]/ps[:, group==1]) / np.sum(Z[:, group==1]/p[:, group==1])
    # qC = np.sum(Zs[:, group==0]/ps[:, group==0]) / np.sum(Z[:, group==0]/p[:, group==0])
    qnum = np.min([qT, qC])
    qden = np.max([qT, qC])
    q = qnum / qden
    g = 0 if qT < qC else 1
    return q, g, qnum


def ns_for_variance(S, Z, p, group, g):
    """
    Calculate the trimming probability q and the trimmed-group indicator g.

    Parameters
    ----------
    S : array 
        n x 1 boolean array indicating the selected sample
    Z : array
        n x t exposure mapping matrix
    p : array
        n x t propensity score matrix
    group : array
        t x1 boolean array indicating the group exposed (1) or control (0)
    """
    Zs = Z[S]
    ps = p[S]
    W = 1 / np.sum(Z*p, axis=1)
    Ws = 1 / np.sum(Zs*ps, axis=1)
    n1 = W[Z[:, group==1].sum(axis=1) == 1].sum() / W.sum() * Z.shape[0]
    n0 = W[Z[:, group==0].sum(axis=1) == 1].sum() / W.sum() * Z.shape[0]
    ns1 = Ws[Zs[:, group==g].sum(axis=1) == 1].sum() / Ws.sum() * Zs.shape[0]
    ns0 = Ws[Zs[:, group!=g].sum(axis=1) == 1].sum() / Ws.sum() * Zs.shape[0]
    return [n1, n0, ns1, ns0]


def trim_quantile(Y, Z, group, p_hat, g):
    """
    Calculate the trimming indicators [lb, ub] for lower and upper bound. 
    Assign 1 when the observation stays, -1 if the observation is trimmed.

    Parameters
    ----------
    Y : array
        Outcome vector.
    Z : array
        (n x 2) treatment matrix.
    group : array
        Boolean array indicating the group exposed.
    p_hat : float
        Trimming probability.
    g : int
        Trimmed-group indicator.
    """
    # Sort the outcome Y in ascending order only for the group g
    trim_ind = Z[:, group==g].sum(axis=1) == 1
    sortidx = np.argsort(Y[trim_ind])
    # Calculate the index at the threshold for trimming
    q_lb = int(np.floor((p_hat * sortidx.size))) if g == 1 else int(np.floor(((1-p_hat) * sortidx.size)))
    q_ub = int(np.floor(((1-p_hat) * sortidx.size))) if g == 1 else int(np.floor((p_hat * sortidx.size)))
    # Trimming indicators: 1 stays, -1 drops
    lb = np.full(Y[trim_ind].size, -1)
    ub = np.full(Y[trim_ind].size, -1)
    if g == 1:
        lb[sortidx[:q_lb]] = 1
        ub[sortidx[q_ub+1:]] = 1
    else:
        lb[sortidx[q_lb+1:]] = 1
        ub[sortidx[:q_ub]] = 1
    trim_lb, trim_ub = trim_ind.copy()*1, trim_ind.copy()*1
    trim_lb[trim_ind], trim_ub[trim_ind] = lb, ub
    y_q = np.quantile(Y[trim_ind], p_hat)
    y_1q = np.quantile(Y[trim_ind], 1-p_hat)
    return [trim_lb, trim_ub], [y_q, y_1q]



# https://setuptools.pypa.io/en/latest/setuptools.html

from setuptools import setup, find_packages

from spillover_effects import __version__

setup(
    name='spillover-effects',
    version=__version__,

    url='https://github.com/pabloestradac/spillover-effects',
    author='Pablo Estrada',
    author_email='pabloestradace@gmail.com',

    packages=find_packages(),

    install_requires=[
        'numpy',
        'pandas',
        'scipy',
        ],
)